@@ -4,82 +4,55 @@
  * SPDX-License-Identifier: MIT
  */
 
-//"use strict";
+// "use strict";
 
 // Various namespace-like modules
 
-var STACK_ALIGN = 16;
-
-var Variables = {
-  globals: {},
-  indexedGlobals: {}, // for indexed globals, ident ==> index
-  // Used in calculation of indexed globals
-  nextIndexedOffset: 0,
-
-  resolveAliasToIdent: function(ident) {
-    while (1) {
-      var varData = Variables.globals[ident];
-      if (!(varData && varData.targetIdent)) break;
-      ident = varData.targetIdent; // might need to eval to turn (6) into 6
-    }
-    return ident;
-  },
-};
-
-var Types = {
-  types: {},
-};
-
 // Constructs an array ['a0', 'a1', 'a2', ..., 'a(n-1)']
 function genArgSequence(n) {
-  var args = [];
-  for(var i = 0; i < n; ++i) {
-    args.push('a'+i);
+  const args = [];
+  for (let i = 0; i < n; ++i) {
+    args.push('a' + i);
   }
   return args;
 }
 
-var Functions = {
-  // All functions that will be implemented in this file. Maps id to signature
-  implementedFunctions: {},
-  // functions added from the library. value 2 means asmLibraryFunction
-  libraryFunctions: {},
-};
+// List of functions that were added from the library.
+global.libraryFunctions = [];
 
-var LibraryManager = {
+global.LibraryManager = {
   library: null,
   structs: {},
   loaded: false,
   libraries: [],
 
   has: function(name) {
-    return this.libraries.indexOf(name) >= 0;
+    return this.libraries.includes(name);
   },
 
   load: function() {
     if (this.library) return;
 
     // Core system libraries (always linked against)
-    var libraries = [
+    let libraries = [
       'library.js',
-      'library_stack.js',
       'library_formatString.js',
       'library_math.js',
       'library_path.js',
-      'library_signals.js',
       'library_syscall.js',
       'library_html5.js',
       'library_stack_trace.js',
       'library_wasi.js',
       'library_int53.js',
-      'library_dylink.js'
+      'library_dylink.js',
+      'library_eventloop.js',
     ];
 
-    if (!EXCEPTION_HANDLING) {
-      if (!DISABLE_EXCEPTION_THROWING) {
-        libraries.push('library_exceptions.js');
-      } else {
+    if (LINK_AS_CXX && !EXCEPTION_HANDLING) {
+      if (DISABLE_EXCEPTION_THROWING) {
         libraries.push('library_exceptions_stub.js');
+      } else {
+        libraries.push('library_exceptions.js');
       }
     }
 
@@ -91,6 +64,7 @@ var LibraryManager = {
       libraries.push('library_strings.js');
     } else {
       libraries.push('library_browser.js');
+      libraries.push('library_wget.js');
     }
 
     if (USE_PTHREADS) { // TODO: Currently WebGL proxying makes pthreads library depend on WebGL.
@@ -98,6 +72,10 @@ var LibraryManager = {
       libraries.push('library_html5_webgl.js');
     }
 
+    if (EMSCRIPTEN_TRACING) {
+      libraries.push('library_memoryprofiler.js');
+    }
+
     if (FILESYSTEM) {
       // Core filesystem libraries (always linked against, unless -s FILESYSTEM=0 is specified)
       libraries = libraries.concat([
@@ -110,11 +88,15 @@ var LibraryManager = {
 
       if (NODERAWFS) {
         // NODERAWFS requires NODEFS
-        if (SYSTEM_JS_LIBRARIES.indexOf('library_nodefs.js') < 0) {
+        if (!JS_LIBRARIES.includes('library_nodefs.js')) {
           libraries.push('library_nodefs.js');
         }
         libraries.push('library_noderawfs.js');
       }
+    } else if (WASMFS) {
+      libraries.push('library_wasmfs.js');
+      libraries.push('library_wasmfs_js_file.js');
+      libraries.push('library_wasmfs_fetch.js');
     }
 
     // Additional JS libraries (without AUTO_JS_LIBRARIES, link to these explicitly via -lxxx.js)
@@ -131,7 +113,7 @@ var LibraryManager = {
         'library_uuid.js',
         'library_glew.js',
         'library_idbstore.js',
-        'library_async.js'
+        'library_async.js',
       ]);
     } else {
       if (ASYNCIFY) {
@@ -145,17 +127,21 @@ var LibraryManager = {
       }
     }
 
-    // Add any explicitly specified system JS libraries to link to, add those to link.
-    libraries = libraries.concat(SYSTEM_JS_LIBRARIES)
-
     if (LZ4) {
       libraries.push('library_lz4.js');
     }
 
     if (MAX_WEBGL_VERSION >= 2) {
+      // library_webgl2.js must be included only after library_webgl.js, so if we are
+      // about to include library_webgl2.js, first squeeze in library_webgl.js.
+      libraries.push('library_webgl.js');
       libraries.push('library_webgl2.js');
     }
 
+    if (GL_EXPLICIT_UNIFORM_LOCATION || GL_EXPLICIT_UNIFORM_BINDING) {
+      libraries.push('library_c_preprocessor.js');
+    }
+
     if (LEGACY_GL_EMULATION) {
       libraries.push('library_glemu.js');
     }
@@ -168,91 +154,128 @@ var LibraryManager = {
     if (BOOTSTRAPPING_STRUCT_INFO) {
       libraries = [
         'library_bootstrap.js',
-        'library_stack.js',
         'library_formatString.js',
-        'library_stack_trace.js'
+        'library_int53.js',
       ];
     }
 
+    if (SUPPORT_BIG_ENDIAN) {
+      libraries.push('library_little_endian_heap.js');
+    }
+
+    // Add all user specified --js-library files to the link.
+    // These must be added last after all Emscripten-provided system libraries
+    // above, so that users can override built-in JS library symbols in their
+    // own code.
+    libraries = libraries.concat(JS_LIBRARIES);
+
     // Deduplicate libraries to avoid processing any library file multiple times
-    libraries = libraries.filter(function(item, pos) {
-      return libraries.indexOf(item) == pos;
-    });
+    libraries = libraries.filter((item, pos) => libraries.indexOf(item) == pos);
 
     // Save the list for has() queries later.
     this.libraries = libraries;
 
-    for (var filename of libraries) {
-      var src = read(filename);
-      var processed = undefined;
+    for (const filename of libraries) {
+      const isUserLibrary = nodePath.isAbsolute(filename);
+      if (VERBOSE) {
+        if (isUserLibrary) {
+          printErr('processing user library: ' + filename);
+        } else {
+          printErr('processing system library: ' + filename);
+        }
+      }
+      const src = read(filename);
+      let origLibrary = undefined;
+      let processed = undefined;
+      // When we parse user libraries also set `__user` attribute
+      // on each element so that we can distinguish them later.
+      if (isUserLibrary) {
+        origLibrary = this.library;
+        this.library = new Proxy(this.library, {
+          set: (target, prop, value) => {
+            target[prop] = value;
+            if (!isJsLibraryConfigIdentifier(prop)) {
+              target[prop + '__user'] = true;
+            }
+            return true;
+          },
+        });
+      }
       try {
         processed = processMacros(preprocess(src, filename));
         eval(processed);
-      } catch(e) {
-        var details = [e, e.lineNumber ? 'line number: ' + e.lineNumber : ''];
+      } catch (e) {
+        const details = [e, e.lineNumber ? `line number: ${e.lineNumber}` : ''];
         if (VERBOSE) {
-          details.push((e.stack || "").toString().replace('Object.<anonymous>', filename));
+          details.push((e.stack || '').toString().replace('Object.<anonymous>', filename));
         }
         if (processed) {
-          error('failure to execute js library "' + filename + '": ' + details);
+          error(`failure to execute js library "${filename}": ${details}`);
           if (VERBOSE) {
-            error('preprocessed source (you can run a js engine on this to get a clearer error message sometimes):\n=============\n' + processed + '\n=============');
+            error(`preprocessed source (you can run a js engine on this to get a clearer error message sometimes):\n=============\n${processed}\n=============`);
           } else {
-            error('use -s VERBOSE to see more details')
+            error('use -s VERBOSE to see more details');
           }
         } else {
-          error('failure to process js library "' + filename + '": ' + details);
+          error(`failure to process js library "${filename}": ${details}`);
           if (VERBOSE) {
-            error('original source:\n=============\n' + src + '\n=============');
+            error(`original source:\n=============\n${src}\n=============`);
           } else {
-            error('use -s VERBOSE to see more details')
+            error('use -s VERBOSE to see more details');
           }
         }
         throw e;
+      } finally {
+        if (origLibrary) {
+          this.library = origLibrary;
+        }
       }
     }
 
     // apply synonyms. these are typically not speed-sensitive, and doing it
     // this way makes it possible to not include hacks in the compiler
     // (and makes it simpler to switch between SDL versions, fastcomp and non-fastcomp, etc.).
-    var lib = LibraryManager.library;
-    libloop: for (var x in lib) {
+    const lib = this.library;
+    libloop: for (const x in lib) {
+      if (!Object.prototype.hasOwnProperty.call(lib, x)) {
+        continue;
+      }
       if (isJsLibraryConfigIdentifier(x)) {
-        var index = x.lastIndexOf('__');
-        var basename = x.slice(0, index);
+        const index = x.lastIndexOf('__');
+        const basename = x.slice(0, index);
         if (!(basename in lib)) {
-          error('Missing library element `' + basename + '` for library config `' + x + '`');
+          error(`Missing library element '${basename}' for library config '${x}'`);
         }
         continue;
       }
-      if (typeof lib[x] === 'string') {
-        var target = x;
-        while (typeof lib[target] === 'string') {
+      if (typeof lib[x] == 'string') {
+        let target = x;
+        while (typeof lib[target] == 'string') {
           // ignore code and variable assignments, aliases are just simple names
           if (lib[target].search(/[=({; ]/) >= 0) continue libloop;
           target = lib[target];
         }
         if (!isNaN(target)) continue; // This is a number, and so cannot be an alias target.
-        if (typeof lib[target] === 'undefined' || typeof lib[target] === 'function') {
+        if (typeof lib[target] == 'undefined' || typeof lib[target] == 'function') {
           // When functions are aliased, a signature for the function must be
           // provided so that an efficient form of forwarding can be
           // implemented.
           function testStringType(sig) {
-            if (typeof lib[sig] !== 'undefined' && typeof typeof lib[sig] !== 'string') {
-              error(sig + ' should be a string! (was ' + typeof lib[sig]);
+            if (typeof lib[sig] != 'undefined' && typeof typeof lib[sig] != 'string') {
+              error(`${sig} should be a string! (was ${typeof lib[sig]})`);
             }
           }
-          var aliasSig = x + '__sig';
-          var targetSig = target + '__sig';
+          const aliasSig = x + '__sig';
+          const targetSig = target + '__sig';
           testStringType(aliasSig);
           testStringType(targetSig);
-          if (typeof lib[aliasSig] === 'string' && typeof lib[targetSig] === 'string' && lib[aliasSig] != lib[targetSig]) {
-            error(aliasSig + ' (' + lib[aliasSig] + ')  differs from ' + targetSig + ' (' + lib[targetSig] + ')');
+          if (typeof lib[aliasSig] == 'string' && typeof lib[targetSig] == 'string' && lib[aliasSig] != lib[targetSig]) {
+            error(`${aliasSig} (${lib[aliasSig]}) differs from ${targetSig} (${lib[targetSig]})`);
           }
 
-          var sig = lib[aliasSig] || lib[targetSig];
-          if (typeof sig !== 'string') {
-            error('Function ' + x + ' aliases to target function ' + target + ', but neither the alias or the target provide a signature. Please add a ' + targetSig + ": 'vifj...' annotation or a " + aliasSig + ": 'vifj...' annotation to describe the type of function forwarding that is needed!");
+          const sig = lib[aliasSig] || lib[targetSig];
+          if (typeof sig != 'string') {
+            error(`Function ${x} aliases to target function ${target}, but neither the alias or the target provide a signature. Please add a ${targetSig}: 'vifj...' annotation or a ${aliasSig}: 'vifj...' annotation to describe the type of function forwarding that is needed!`);
           }
 
           // If only one of the target or the alias specifies a sig then copy
@@ -263,17 +286,17 @@ var LibraryManager = {
             lib[targetSig] = lib[aliasSig];
           }
 
-          if (typeof lib[target] !== 'function') {
+          if (typeof lib[target] != 'function') {
             error(`no alias found for ${x}`);
           }
 
-          var argCount = sig.length - 1;
+          const argCount = sig.length - 1;
           if (argCount !== lib[target].length) {
             error(`incorrect number of arguments in signature of ${x} (declared: ${argCount}, expected: ${lib[target].length})`);
           }
-          var ret = sig == 'v' ? '' : 'return ';
-          var args = genArgSequence(argCount).join(',');
-          lib[x] = new Function(args, ret + '_' + target + '(' + args + ');');
+          const ret = sig == 'v' ? '' : 'return ';
+          const args = genArgSequence(argCount).join(',');
+          lib[x] = new Function(args, `${ret}_${target}(${args});`);
 
           if (!lib[x + '__deps']) lib[x + '__deps'] = [];
           lib[x + '__deps'].push(target);
@@ -283,31 +306,11 @@ var LibraryManager = {
 
     this.loaded = true;
   },
-
-  // Given an ident, see if it is an alias for something, and so forth, returning
-  // the earliest ancestor (the root)
-  getRootIdent: function(ident) {
-    if (!this.library) return null;
-    var ret = LibraryManager.library[ident];
-    if (!ret) return null;
-    var last = ident;
-    while (typeof ret === 'string') {
-      last = ret;
-      ret = LibraryManager.library[ret];
-    }
-    return last;
-  },
-
-  isStubFunction: function(ident) {
-    var libCall = LibraryManager.library[ident.substr(1)];
-    return typeof libCall === 'function' && libCall.toString().replace(/\s/g, '') === 'function(){}'
-                                         && !(ident in Functions.implementedFunctions);
-  }
 };
 
 if (!BOOTSTRAPPING_STRUCT_INFO) {
   // Load struct and define information.
-  var temp = JSON.parse(read(STRUCT_INFO));
+  const temp = JSON.parse(read(STRUCT_INFO));
   C_STRUCTS = temp.structs;
   C_DEFINES = temp.defines;
 } else {
@@ -317,14 +320,10 @@ if (!BOOTSTRAPPING_STRUCT_INFO) {
 
 // Safe way to access a C define. We check that we don't add library functions with missing defines.
 function cDefine(key) {
-	if (key in C_DEFINES) return C_DEFINES[key];
-	throw 'Missing C define ' + key + '! If you just added it to struct_info.json, you need to ./emcc --clear-cache';
+  if (key in C_DEFINES) return C_DEFINES[key];
+  throw new Error(`Missing C define ${key}! If you just added it to struct_info.json, you need to ./emcc --clear-cache`);
 }
 
-var EXPORTED_RUNTIME_METHODS_SET = set(EXPORTED_RUNTIME_METHODS.concat(EXTRA_EXPORTED_RUNTIME_METHODS));
-EXPORTED_RUNTIME_METHODS = unset(EXPORTED_RUNTIME_METHODS_SET);
-EXTRA_EXPORTED_RUNTIME_METHODS = [];
-
 function isFSPrefixed(name) {
   return name.length > 3 && name[0] === 'F' && name[1] === 'S' && name[2] === '_';
 }
@@ -343,14 +342,16 @@ function isExportedByForceFilesystem(name) {
 
 // export parts of the JS runtime that the user asked for
 function exportRuntime() {
+  const EXPORTED_RUNTIME_METHODS_SET = new Set(EXPORTED_RUNTIME_METHODS);
+
   // optionally export something.
   // in ASSERTIONS mode we show a useful error if it is used without
   // being exported. how we show the message depends on whether it's
   // a function (almost all of them) or a number.
   function maybeExport(name, isNumber) {
     // if requested to be exported, export it
-    if (name in EXPORTED_RUNTIME_METHODS_SET) {
-      var exported = name;
+    if (EXPORTED_RUNTIME_METHODS_SET.has(name)) {
+      let exported = name;
       // the exported name may differ from the internal name
       if (isFSPrefixed(exported)) {
         // this is a filesystem value, FS.x exported as FS_x
@@ -360,7 +361,7 @@ function exportRuntime() {
       } else if (exported === 'printErr') {
         exported = 'err';
       }
-      return 'Module["' + name + '"] = ' + exported + ';';
+      return `Module["${name}"] = ${exported};`;
     }
     // do not export it. but if ASSERTIONS, emit a
     // stub with an error, so the user gets a message
@@ -369,14 +370,14 @@ function exportRuntime() {
       // check if it already exists, to support EXPORT_ALL and other cases
       // (we could optimize this, but in ASSERTIONS mode code size doesn't
       // matter anyhow)
-      var extra = '';
+      let extra = '';
       if (isExportedByForceFilesystem(name)) {
         extra = '. Alternatively, forcing filesystem support (-s FORCE_FILESYSTEM=1) can export this for you';
       }
       if (!isNumber) {
-        return 'if (!Object.getOwnPropertyDescriptor(Module, "' + name + '")) Module["' + name + '"] = function() { abort("\'' + name + '\' was not exported. add it to EXTRA_EXPORTED_RUNTIME_METHODS (see the FAQ)' + extra + '") };';
+        return `if (!Object.getOwnPropertyDescriptor(Module, "${name}")) Module["${name}"] = () => abort("'${name}' was not exported. add it to EXPORTED_RUNTIME_METHODS (see the FAQ)${extra}");`;
       } else {
-        return 'if (!Object.getOwnPropertyDescriptor(Module, "' + name + '")) Object.defineProperty(Module, "' + name + '", { configurable: true, get: function() { abort("\'' + name + '\' was not exported. add it to EXTRA_EXPORTED_RUNTIME_METHODS (see the FAQ)' + extra + '") } });';
+        return `if (!Object.getOwnPropertyDescriptor(Module, "${name}")) Object.defineProperty(Module, "${name}", { configurable: true, get: function() { abort("'${name}' was not exported. add it to EXPORTED_RUNTIME_METHODS (see the FAQ)${extra}") } });`;
       }
     }
     return '';
@@ -386,8 +387,8 @@ function exportRuntime() {
     return maybeExport(name, true);
   }
 
-  // All possible runtime elements to export
-  var runtimeElements = [
+  // All possible runtime elements that can be exported
+  let runtimeElements = [
     'intArrayFromString',
     'intArrayToString',
     'ccall',
@@ -427,7 +428,6 @@ function exportRuntime() {
     'removeFunction',
     'getFuncWrapper',
     'prettyPrint',
-    'makeBigInt',
     'dynCall',
     'getCompilerSetting',
     'print',
@@ -436,11 +436,20 @@ function exportRuntime() {
     'setTempRet0',
     'callMain',
     'abort',
+    'keepRuntimeAlive',
   ];
 
+  if (USE_OFFSET_CONVERTER) {
+    runtimeElements.push('WasmOffsetConverter');
+  }
+
+  if (LOAD_SOURCE_MAP) {
+    runtimeElements.push('WasmSourceMap');
+  }
+
   // Add JS library elements such as FS, GL, ENV, etc. These are prefixed with
   // '$ which indicates they are JS methods.
-  for (var ident in LibraryManager.library) {
+  for (const ident in LibraryManager.library) {
     if (ident[0] === '$' && !isJsLibraryConfigIdentifier(ident)) {
       runtimeElements.push(ident.substr(1));
     }
@@ -462,9 +471,8 @@ function exportRuntime() {
       'stringToUTF32',
       'lengthBytesUTF32',
       'allocateUTF8',
-      'allocateUTF8OnStack'
+      'allocateUTF8OnStack',
     ]);
-
   }
 
   if (STACK_OVERFLOW_CHECK) {
@@ -476,13 +484,13 @@ function exportRuntime() {
     // In pthreads mode, the following functions always need to be exported to
     // Module for closure compiler, and also for MODULARIZE (so worker.js can
     // access them).
-    var threadExports = ['PThread', 'wasmMemory'];
+    const threadExports = ['PThread', 'wasmMemory'];
     if (!MINIMAL_RUNTIME) {
       threadExports.push('ExitStatus');
     }
 
-    threadExports.forEach(function(x) {
-      EXPORTED_RUNTIME_METHODS_SET[x] = 1;
+    threadExports.forEach((x) => {
+      EXPORTED_RUNTIME_METHODS_SET.add(x);
       runtimeElements.push(x);
     });
   }
@@ -494,59 +502,26 @@ function exportRuntime() {
   // dynCall_* methods are not hardcoded here, as they
   // depend on the file being compiled. check for them
   // and add them.
-  for (var name in EXPORTED_RUNTIME_METHODS_SET) {
+  for (const name of EXPORTED_RUNTIME_METHODS_SET) {
     if (/^dynCall_/.test(name)) {
       // a specific dynCall; add to the list
       runtimeElements.push(name);
     }
   }
-  var runtimeNumbers = [
+  const runtimeNumbers = [
     'ALLOC_NORMAL',
     'ALLOC_STACK',
   ];
   if (ASSERTIONS) {
     // check all exported things exist, warn about typos
-    for (var name in EXPORTED_RUNTIME_METHODS_SET) {
-      if (runtimeElements.indexOf(name) < 0 &&
-          runtimeNumbers.indexOf(name) < 0) {
-        printErr('warning: invalid item (maybe a typo?) in EXPORTED_RUNTIME_METHODS: ' + name);
+    for (const name of EXPORTED_RUNTIME_METHODS_SET) {
+      if (!runtimeElements.includes(name) && !runtimeNumbers.includes(name)) {
+        printErr(`warning: invalid item (maybe a typo?) in EXPORTED_RUNTIME_METHODS: ${name}`);
       }
     }
   }
-  var exports = runtimeElements.map(function(name) { return maybeExport(name); });
-  exports = exports.concat(runtimeNumbers.map(function(name) { return maybeExportNumber(name); }));
-  exports = exports.filter(function(name) { return name != '' });
+  let exports = runtimeElements.map((name) => maybeExport(name));
+  exports = exports.concat(runtimeNumbers.map((name) => maybeExportNumber(name)));
+  exports = exports.filter((name) => name != '');
   return exports.join('\n');
 }
-
-var PassManager = {
-  serialize: function() {
-    print('\n//FORWARDED_DATA:' + JSON.stringify({
-      Functions: Functions,
-      EXPORTED_FUNCTIONS: EXPORTED_FUNCTIONS,
-      ATINITS: ATINITS.join('\n'),
-      ATMAINS: ATMAINS.join('\n'),
-      ATEXITS: ATEXITS.join('\n'),
-    }));
-  },
-  load: function(json) {
-    var data = JSON.parse(json);
-    for (var i in data.Types) {
-      Types[i] = data.Types[i];
-    }
-    for (var i in data.Variables) {
-      Variables[i] = data.Variables[i];
-    }
-    for (var i in data.Functions) {
-      Functions[i] = data.Functions[i];
-    }
-    EXPORTED_FUNCTIONS = data.EXPORTED_FUNCTIONS;
-    /*
-    print('\n//LOADED_DATA:' + JSON.stringify({
-      Types: Types,
-      Variables: Variables,
-      Functions: Functions
-    }));
-    */
-  }
-};
