@@ -6,7 +6,7 @@
 
 #include "runtime_safe_heap.js"
 
-#if ASSERTIONS
+#if ASSERTIONS || SAFE_HEAP
 /** @type {function(*, string=)} */
 function assert(condition, text) {
   if (!condition) throw text;
@@ -29,12 +29,8 @@ var tempI64;
 #endif
 
 var tempRet0 = 0;
-var setTempRet0 = function(value) {
-  tempRet0 = value;
-}
-var getTempRet0 = function() {
-  return tempRet0;
-}
+var setTempRet0 = (value) => { tempRet0 = value };
+var getTempRet0 = () => tempRet0;
 
 function alignUp(x, multiple) {
   if (x % multiple > 0) {
@@ -64,11 +60,18 @@ Module['wasm'] = base64Decode('<<< WASM_BINARY_DATA >>>');
 var HEAP8, HEAP16, HEAP32, HEAPU8, HEAPU16, HEAPU32, HEAPF32, HEAPF64;
 var wasmMemory, buffer, wasmTable;
 
+#if SUPPORT_BIG_ENDIAN
+var HEAP_DATA_VIEW;
+#endif
+
 function updateGlobalBufferAndViews(b) {
 #if ASSERTIONS && USE_PTHREADS
   assert(b instanceof SharedArrayBuffer, 'requested a shared WebAssembly.Memory but the returned buffer is not a SharedArrayBuffer, indicating that while the browser has SharedArrayBuffer it does not have WebAssembly threads support - you may need to set a flag');
 #endif
   buffer = b;
+#if SUPPORT_BIG_ENDIAN
+  HEAP_DATA_VIEW = new DataView(b);
+#endif
   HEAP8 = new Int8Array(b);
   HEAP16 = new Int16Array(b);
   HEAP32 = new Int32Array(b);
@@ -83,14 +86,14 @@ function updateGlobalBufferAndViews(b) {
 #if USE_PTHREADS
 if (!ENVIRONMENT_IS_PTHREAD) {
 #endif
-#if ALLOW_MEMORY_GROWTH && MAXIMUM_MEMORY != -1
+#if ALLOW_MEMORY_GROWTH && MAXIMUM_MEMORY != FOUR_GB
   var wasmMaximumMemory = {{{ MAXIMUM_MEMORY >>> 16 }}};
 #else
   var wasmMaximumMemory = {{{ INITIAL_MEMORY >>> 16}}};
 #endif
   wasmMemory = new WebAssembly.Memory({
     'initial': {{{ INITIAL_MEMORY >>> 16 }}}
-#if USE_PTHREADS || !ALLOW_MEMORY_GROWTH || MAXIMUM_MEMORY != -1
+#if USE_PTHREADS || !ALLOW_MEMORY_GROWTH || MAXIMUM_MEMORY != FOUR_GB
     , 'maximum': wasmMaximumMemory
 #endif
 #if USE_PTHREADS
@@ -101,7 +104,7 @@ if (!ENVIRONMENT_IS_PTHREAD) {
 #if USE_PTHREADS
 } else {
 #if MODULARIZE
-  updateGlobalBufferAndViews({{{EXPORT_NAME}}}.buffer);
+  updateGlobalBufferAndViews(Module.buffer);
 #else
   updateGlobalBufferAndViews(wasmMemory.buffer);
 #endif
@@ -126,7 +129,7 @@ var wasmOffsetConverter;
 var __ATEXIT__    = []; // functions called during shutdown
 #endif
 
-#if ASSERTIONS || SAFE_HEAP
+#if ASSERTIONS || SAFE_HEAP || USE_ASAN
 var runtimeInitialized = false;
 
 // This is always false in minimal_runtime - the runtime does not have a concept
