@@ -1,11 +1,10 @@
 #include <stdio.h>
 #include <fcntl.h>
+#include <stdlib.h>
 #include "stdio_impl.h"
 
 #define MAXTRIES 100
 
-char *__randname(char *);
-
 FILE *tmpfile(void)
 {
 	char s[] = "/tmp/tmpfile_XXXXXX";
@@ -33,4 +32,4 @@ FILE *tmpfile(void)
 	return 0;
 }
 
-LFS64(tmpfile);
+weak_alias(tmpfile, tmpfile64);
