@@ -1,11 +1,10 @@
 #ifdef __EMSCRIPTEN__
 #include <math.h>
+#include <emscripten/threading.h>
 #endif
 
 #include "pthread_impl.h"
 
-int _pthread_isduecanceled(struct pthread *pthread_ptr);
-
 void __wait(volatile int *addr, volatile int *waiters, int val, int priv)
 {
 	int spins=100;
@@ -16,22 +15,24 @@ void __wait(volatile int *addr, volatile int *waiters, int val, int priv)
 	}
 	if (waiters) a_inc(waiters);
 #ifdef __EMSCRIPTEN__
-	int is_main_thread = emscripten_is_main_runtime_thread();
+	int is_runtime_thread = emscripten_is_main_runtime_thread();
+
+	// Main runtime thread may need to run proxied calls, so sleep in very small slices to be responsive.
+	double max_ms_slice_to_sleep = is_runtime_thread ? 1 : 100;
+
 	while (*addr==val) {
-		if (pthread_self()->cancelasync == PTHREAD_CANCEL_ASYNCHRONOUS) {
-			// Must wait in slices in case this thread is cancelled in between.
+		if (is_runtime_thread || pthread_self()->cancelasync == PTHREAD_CANCEL_ASYNCHRONOUS) {
 			int e;
 			do {
-				if (_pthread_isduecanceled(pthread_self())) {
+				if (pthread_self()->cancel) {
 					if (waiters) a_dec(waiters);
 					return;
 				}
 				// Assist other threads by executing proxied operations that are effectively singlethreaded.
-				if (is_main_thread) emscripten_main_thread_process_queued_calls();
-				// Main thread waits in _very_ small slices so that it stays responsive to assist proxied
-				// pthread calls.
-				e = emscripten_futex_wait((void*)addr, val, is_main_thread ? 1 : 100);
-			} while(e == -ETIMEDOUT);
+				if (is_runtime_thread) emscripten_main_thread_process_queued_calls();
+				// Must wait in slices in case this thread is cancelled in between.
+				e = emscripten_futex_wait((void*)addr, val, max_ms_slice_to_sleep);
+			} while (e == -ETIMEDOUT);
 		} else {
 			// Can wait in one go.
 			emscripten_futex_wait((void*)addr, val, INFINITY);
