@@ -12,16 +12,22 @@
 
 int main(int argc, char* argv[]) {
   // Alignment check for mmap below should be consistent with the reported page size
-  // For now, just require it to be 16384
-  assert(getpagesize() == 16384);
-  assert(sysconf(_SC_PAGESIZE) == 16384);
+  printf("getpagesize(): %d\n", getpagesize());
+  printf("sysconf(_SC_PAGESIZE): %ld\n", sysconf(_SC_PAGESIZE));
+  assert(getpagesize() == 65536);
+  assert(sysconf(_SC_PAGESIZE) == 65536);
 
+  int* maps[10];
   for (int i = 0; i < 10; i++) {
     int* map = (int*)mmap(0, 5000, PROT_READ | PROT_WRITE,
                           MAP_SHARED | MAP_ANON, -1, 0);
     assert(map != MAP_FAILED);
-    assert(((int)map) % 16384 == 0); // aligned
-    assert(munmap(map, 5000) == 0);
+    assert(((long)map) % 65536 == 0); // aligned
+    maps[i] = map;
+  }
+
+  for (int i = 0; i < 10; i++) {
+    assert(munmap(maps[i], 5000) == 0);
   }
 
   const int NUM_BYTES = 8 * 1024 * 1024;
@@ -43,6 +49,6 @@ int main(int argc, char* argv[]) {
 
   assert(munmap(map, NUM_BYTES) == 0);
 
-  printf("hello,world");
+  printf("hello,world\n");
   return 0;
 }
