@@ -19,7 +19,7 @@ void *thread_main(void *arg)
   globalData = 20;
   EM_ASM(out('hello from pthread 2: ' + $0), globalData);
   assert(globalData == 20);
-	return 0;
+  return 0;
 }
 
 int main()
@@ -31,13 +31,11 @@ int main()
   EM_ASM(out('hello from main 2: ' + $0), globalData);
   assert(globalData == 10);
 
-	pthread_t thread;
-	pthread_create(&thread, NULL, thread_main, NULL);
+  pthread_t thread;
+  pthread_create(&thread, NULL, thread_main, NULL);
   pthread_join(thread, 0);
 
   EM_ASM(out('hello from main 3: ' + $0), globalData);
   assert(globalData == 20);
-#ifdef REPORT_RESULT
-  REPORT_RESULT(globalData);
-#endif
+  return 0;
 }
