@@ -6,8 +6,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <pthread.h>
-#include <emscripten.h>
-#include <emscripten/threading.h>
 #include <errno.h>
 #include <assert.h>
 
@@ -21,18 +19,11 @@ int main()
 {
 	pthread_t thread;
 	int rc = pthread_create(&thread, NULL, ThreadMain, 0);
-	if (emscripten_has_threading_support()) {
-		assert(rc == 0);
+	assert(rc == 0);
 
-		rc = pthread_join(thread, NULL);
-		assert(rc == 0);
+	rc = pthread_join(thread, NULL);
+	assert(rc == 0);
 
-		printf("The thread should print 'Hello from thread, string: str, int: 5, double: 42.0'\n");
-	} else {
-		assert(rc == EAGAIN);
-	}
-
-#ifdef REPORT_RESULT
-	REPORT_RESULT(0);
-#endif
+	printf("The thread should print 'Hello from thread, string: str, int: 5, double: 42.0'\n");
+	return 0;
 }
