@@ -7,15 +7,12 @@
 #include <stdlib.h>
 #include <assert.h>
 
-int main()
-{
+int main() {
 	void *ptr1 = malloc(24*1024*1024);
 	printf("ptr1: %p\n", ptr1);
 	assert(ptr1 != 0);
 	void *ptr2 = malloc(24*1024*1024);
 	printf("ptr2: %p\n", ptr2);
 	assert(ptr2 == 0);
-#ifdef REPORT_RESULT
-	REPORT_RESULT(1);
-#endif
+	return 0;
 }
