@@ -5,6 +5,7 @@
 
 import os
 import shutil
+from pathlib import Path
 
 TAG = 'version_1'
 HASH = '0d0b1280ba0501ad0a23cf1daa1f86821c722218b59432734d3087a89acd22aabd5c3e5e1269700dcd41e87073046e906060f167c032eb91a3ac8c5808a02783'
@@ -26,7 +27,7 @@ def get(ports, settings, shared):
     os.makedirs(dest_path)
     shutil.rmtree(dest_path, ignore_errors=True)
     shutil.copytree(source_path, dest_path)
-    open(os.path.join(dest_path, 'include/ftconfig.h'), 'w').write(ftconf_h)
+    Path(dest_path, 'include/ftconfig.h').write_text(ftconf_h)
 
     # build
     srcs = ['src/autofit/autofit.c',
@@ -111,7 +112,7 @@ def get(ports, settings, shared):
 
 
 def clear(ports, settings, shared):
-  shared.Cache.erase_file('libfreetype.a')
+  shared.Cache.erase_lib('libfreetype.a')
 
 
 def process_args(ports):
